//----------------------------------------------------------------------------
// File:   c:\cbproj\AudioIO\AudioIO_Test_Client.c
// Author: Wolfgang Buescher (DL4YHF)
// Date:   2011-07-05
// Purpose: Test application for the "Audio-To-Winamp" plugin
//           (which is, from Winamp's point of view, an INPUT plugin,
//            but from this application's point of view, it's an AUDIO SINK )
//
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Copyright, License, and Disclaimer
//      Copyright (c) 2008...2065, Wolfgang Bscher (DL4YHF)
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are met:
//   1. Redistributions of source code must retain the above copyright notice,
//      this list of conditions and the following disclaimer.
//   2. Redistributions in binary form must reproduce the above copyright notice,
//      this list of conditions and the following disclaimer in the
//      documentation and/or other materials provided with the distribution.
//   3. Neither the name of the author nor the name of other contributors
//      may be used to endorse or promote products derived from this software
//      without specific prior written permission.
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
//  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
//   OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
//   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
//   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
// STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
// EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//
// Revision history:
//  V0.00, 2008-03-27:  Project started as a WINAMP-"input"-plugin
//                      to send audio data from DL4YHF's "Spectrum Lab"
//                      into Winamp, to play with Icecast/Shoutcast streaming .
//
//----------------------------------------------------------------------------


#include <stdio.h>
#include <conio.h>
#include <math.h>

#include "AudioIO.h"  // data types, error codes, and prototypes for Audio-I/O

#include "utility1.h" // by DL4YHF, with stuff like UTL_ConvertYMDhmsToUnix()

#ifndef  C_PI
# define C_PI 3.141592653589793
#endif

char AioTest_sz40InputStreamID[44] = "Test";
char AioTest_sz40OutputStreamID[44]= "Test";
char AioTest_sz80AudioSrcDescr[84] = "Source description";
char AioTest_sz40AudioReaderID[44] = "AudioIOTestClient";

T_AIO_DLLHostInstanceData AIO_Host;
T_ChunkInfo ChunkInfoForOutput, ChunkInfoFromInput;
HWND hwndControlPanel  = NULL;  // handle of the Audio-I/O-DLL's control panel
BOOL fControlPanelClosed=FALSE;

/****************** Routines outside TDebugForm *****************************/
void DEBUG_EnterErrorHistory(
        int  debug_level,   // importance of the message (to be entered or not)
        int  options,       // reserved for future extensions
     double time_of_occurrence,  // fractional GMT SECONDS since Jan.1970 ,
                           // may be UTL_USE_CURRENT_TIME
       char *message_fmt,  // pointer to null-terminated format string (source)
                   ... )   // optional argument list (printf-style)
  // CAUTION ! THIS ROUTINE MAY BE CALLED FROM SOME "WORKER THREAD" !
  // (thus it must not use VCL-stuff, because Borland's VCL isn't thread-safe)
{
  va_list parameter;       // argument list for VA_... macros
  char buf[512];
  char *cp = buf;

  (void)debug_level;  // not used here (but in Spectrum Lab)
  (void)options;
  (void)time_of_occurrence; // ..

  va_start( parameter, message_fmt );
  vsprintf( cp, message_fmt, parameter );    // format argument list
  va_end  ( parameter );

  printf( "\r\nDEBUG: ERROR: %s\r\n", buf );

} // end DEBUG_EnterErrorHistory()


/***************************************************************************/
void DEBUG_Breakpoint( void ) // common breakpoint for all "self detected errors"
  // Only called through macro DEBUGGER_BREAK() in 'beta releases' !
  // Replaces assert()-stuff .  When running under debugger control,
  // set a breakpoint ON THE LAST CURLY BRACE of the function body .
  // Then use "return to caller", if possible .
{
  // ex: geninterrupt(0x03);      // int 3   ->  debugger break (always)
  // removed; geninterrupt(3) crashed the application when running w/o debugger
} // end DEBUG_Breakpoint() .  ALWAYS SET A BREAKPOINT HERE ! <<<<<<<<<<<<<<<
  // Return to caller to find the cause. We *should* never get here, but...


/***************************************************************************/
INT32 _stdcall AIO_Host_Callback4ControlPanel(
                 void  *pvUserDataForCallback,
                 INT32 i32FunctionCode,
                 INT32 iParam1,
                 INT32 iParam2)
{
  INT32 i32Result = AIO_NO_ERROR;
  (void)pvUserDataForCallback;  // unused parameter (still)
  (void)iParam1; (void)iParam2; // unused parameters
  switch( i32FunctionCode ) // check the 'function code' :
   { case WM_AUDIO_IO_FC_NONE:         /* just a "ping" */
        break;
     case WM_AUDIO_IO_FC_PANEL_CLOSED: /* control panel closed by user */
        fControlPanelClosed = TRUE;
        break;
   } // end switch i32FunctionCode

  return i32Result;
} // end AIO_Host_Callback4ControlPanel()

/***************************************************************************/
BOOL HandleEvents(void)  // handles WINDOWS events ("message pump" for the DLL's control panel)
  // Crude replacement for the application's "message pump"
  // (only for this CONSOLE APPLICATION, never try this in a GUI app)
{
  static BOOL fContinueLoop = TRUE;
  MSG  msg;

  // if( ! fShowControlPanel )    // message pump NOT active when running as a 'normal' console app
  //  { return FALSE;
  //  }
  while( fContinueLoop && PeekMessage(&msg,0,0,0,PM_REMOVE))
   {
     switch(msg.message)
      { case WM_QUIT:   // receiving WM_QUIT terminates *THE APPLICATION*
           fContinueLoop = FALSE;
           // Note: The Audio-I/O-DLL's control panel does NOT (!!!)
           //       send WM_QUIT because WM_QUIT terminates THE APPLICATION.
           break;
        case WM_DESTROY: // "someone" was closed or destroyed, but WHO ?
           if( msg.hwnd==hwndControlPanel )
            { hwndControlPanel = NULL;  // this handle is INVALID after this
              fControlPanelClosed = TRUE;
              fContinueLoop = FALSE;
            }
           break;
        case WM_AUDIO_IO_CONTROL:  // info from the control panel for the host:
           // WM_AUDIO_IO_CONTROL( wParam= function code, lParam= function value)
           switch( msg.wParam ) // check the 'function code' in MESSAGE.wParam :
            { case WM_AUDIO_IO_FC_NONE:         /* just a "ping" */
                 break;
              case WM_AUDIO_IO_FC_PANEL_CLOSED: /* control panel closed by user */
                 fControlPanelClosed = TRUE;
                 break;
            } // end switch wParam, here: "function code"
           break; // end case WM_AUDIO_IO_CONTROL
      }
     TranslateMessage(&msg);
     DispatchMessage(&msg);
   } // end while
  return fContinueLoop;  // possible return values :
    // TRUE = "call me again later; at the moment there's nothing to do"
    // FALSE= "the application should quit"

} // end HandleEvents()



/***************************************************************************/
void OpenAudioOutput(void)  // invoked through the /o command line parameter
{
  long i32Result;
  printf("\r\nAIO_OpenAudioOutput, to transmit stream-ID \"%s\" ..",
                   (char*)AioTest_sz40OutputStreamID );
  if( (i32Result=AIO_Host_OpenAudioOutput(
              &AIO_Host,   // [in,out] DLL-host instance data
              AioTest_sz40OutputStreamID,// [in] unique name identifying the stream
              AioTest_sz80AudioSrcDescr, // [in] source description, see AudioIO.h
              ChunkInfoForOutput.dblPrecSamplingRate, // [in] samplerate (samples/sec)
              2,                         // [in] 1=mono, 2=stereo, 4=quadrophonic
              AIO_TIMEOUT_DEFAULT,       // [in] iTimeout_ms
              AIO_OPEN_OPTION_NORMAL)    // [in] bit combination of AIO_OPEN_OPTION_... flags
            ) < 0 )
   { printf("\r\nFailed to open the audio-output for writing,\r\n\terror code %d = %s .",
                   (int)i32Result, (char*)AIO_Host_ErrorCodeToString(&AIO_Host, i32Result) );
   } // end if < AIO_OpenAudioOutput() ok >
  else
   { printf("..ok.");
   }
} // end OpenAudioOutput()


/***************************************************************************/
void OpenAudioInput(void)  // invoked through the /i command line parameter
{
  long i32Result;
  T_ChunkInfo chunk_info;
  double dblSamplesPerSecond;

  printf("\r\nAIO_OpenAudioInput, to receive stream-ID \"%s\" ..",
                   (char*)AioTest_sz40InputStreamID );
     // Because certain ExtIO-DLL come up with their own stupid MESSAGE BOX
     // when something goes wrong, instead of simply returning an error code,
     // the radio's VFO FREQUENCY must be set to a valid value.
     // At least, this is true for ExtIO_RTL.DLL in combination with a Noxon DAB USB Stick:
  AIO_Host_SetVFOFrequency( &AIO_Host, 434000000 ); // 434 MHz should be ok for the PLL

     // 2012-11-22 : Oh dear. When opening ExtIO_SDR.DLL, text was printed
     // on the CONSOLE WINDOW (stdout? stderr?) by the DLL:
     //     "Found Fitipower FC0013 tuner" ?!
     // This may cause a 'pure GUI' application to crash (so it did, with SpecLab).
     // A DLL should NEVER, NEVER, NEVER print to 'stdout' by its own gusto !
  if( (i32Result=AIO_Host_OpenAudioInput(
              &AIO_Host,                 // [in,out] DLL-host instance data
              AioTest_sz40InputStreamID, // [in] unique name identifying the stream
              AioTest_sz40AudioReaderID, // [in] unique name identifying the reader
              44100.0,                   // [in] "wanted" sample rate, often IGNORED
              &dblSamplesPerSecond,      // [out(!)] realized sample rate
              2,            // [in] number of channels PER SAMPLE POINT
              1000,         // [in] max timeout in milliseconds for THIS call
              // The open-function may have to wait for a short time, for example while...
              //   - the I/O lib communicates with an external device (to open it)
              //   - the I/O lib waits until the 'writer' (=the other side)
              //     has called AIO_Host_OpenAudioOutput() and set the sampling rate.
              // Test BOTH scenarios (with the 'Audio I/O Test Client') :
              //   - writer launched before reader,
              //   - reader launched before writer .
              AIO_OPEN_OPTION_NORMAL,    // [in] bit combination of AIO_OPEN_OPTION_... flags
              &chunk_info)  // [out,optional] chunk info with precise realized sampling rate, etc
            ) < 0 )
   { printf("\r\nFailed to open the audio-input for reading,\r\n\terror code %d = %s .",
                   (int)i32Result, (char*)AIO_Host_ErrorCodeToString(&AIO_Host, i32Result) );
   } // end if < AIO_OpenAudioOutput() ok >
  else
   { printf("..ok, %.1lf S/s", (double)dblSamplesPerSecond );

#  if( SWI_SUPPORT_EXTIO_DLL )   // support Winrad-compatible ExtIO-DLLs ?
     if( AIO_Host.fExtIOLoaded )
      { printf("\r\nDLL identified as ExtIO-DLL with these properties:\r\n");
        printf("  DevName=\"%s\", HWName=\"%s\", HWModel=\"%s\" .\r\n",
               AIO_Host.devNameIn, AIO_Host.extHWname, AIO_Host.extHWmodel );
        printf("  HasVFO=%s, HasControlPanel=%s, IQpairsPerCallback=%ld .",
               AIO_Host_HasVFO(&AIO_Host)          ? "YES" : "no",
               AIO_Host_HasControlPanel(&AIO_Host) ? "YES" : "no",
               (long)AIO_Host.nIQPairsPerExtIoCallback  );
      }
#  endif // SWI_SUPPORT_EXTIO_DLL ?
   } // end else < successfully opened the whatever-it-is for INPUT >
} // end OpenAudioInput()


/***************************************************************************/
void DumpChunkInfo( T_ChunkInfo *pChunkInfo )
{
  int year,month,day,hour,minute;
  double dblSeconds;
  printf("\r\nAudio Chunk Info: %ld samples/chunk, %lg samples/sec,  %d chn/sample,",
              (long)pChunkInfo->dwNrOfSamplePoints,
            (double)pChunkInfo->dblPrecSamplingRate,
               (int)pChunkInfo->nChannelsPerSample );
  printf("\r\n  total samples=%ld, sizeof(ChunkInfo)=%d, range=%lg, validity=%lX .",
              (long)pChunkInfo->i64TotalSampleCounter,
               (int)pChunkInfo->dwSizeOfStruct,
            (double)pChunkInfo->dblSampleValueRange,
              (long)pChunkInfo->dwValidityFlags );
  // ex: (removed; the 'sample format' must ALWAYS be valid)
  //if( pChunkInfo->dwValidityFlags & CHUNK_INFO_SAMPLE_FORMAT_VALID    )
  // {
  // }
  if( pChunkInfo->dwValidityFlags & CHUNK_INFO_SAMPLE_RATE_VALID      )
   {
   }
  if( pChunkInfo->dwValidityFlags & CHUNK_INFO_SAMPLE_RATE_CALIBRATED )
   {
   }
  if( pChunkInfo->dwValidityFlags & CHUNK_INFO_TIMESTAMPS_VALID       )
   { UTL_SplitUnixDateTimeToYMDhms( pChunkInfo->ldblUnixDateAndTime,
         &year, &month, &day, &hour, &minute, &dblSeconds );
     printf("\r\n  %04d-%02d-%02d %02d:%02d:%02d.%02d",
          year, month, day, hour, minute, (int)dblSeconds, (int)(fmod(dblSeconds*100.0,100.0)) );
   }

  printf("\r\n");
} // end DumpChunkInfo()


/***************************************************************************/
#ifdef __BORLANDC__  // Borland wants to have main() in a file managed by the ID
int main2(int argc, char* argv[]) // (we don't, so all we do in main is call main2)
#else                // Not Borland / C++Builder but (most likely) DevCpp :
int main(int argc, char* argv[])
#endif // not Borland
  // Parameter (example) :  c:\cbproj\AudioIO\in_AudioIO_BCB4_DONT_USE.dll
  //
{
  #define L_SAMPLES_PER_BUFFER 4096
  #define L_CHANNELS_PER_BUFFER 2
  float fltOutputBuffer[L_SAMPLES_PER_BUFFER * L_CHANNELS_PER_BUFFER];
  float fltInputBuffer[ L_SAMPLES_PER_BUFFER * L_CHANNELS_PER_BUFFER];
  int  i,j;
  long i32Result, nSamplesRead;
  int  input_error, output_error;
  double dblWinampSampleRate;
  double dblPhaseAccu = 0.0;
  double dblFreqMod, dblFMPhaseAccu = 0.0;
  double dblPhaseAccuInc, dblFreqModPhaseAccuInc;
  double dblFactor;
  double dblChunkLengthInSeconds;
  int  iWinampNumChannels, iWinampInitialized, iWinampInputPaused;
  INT32 iHaveWaited_ms, iHadWaited_ms; // optional "been-waiting-flag" (32-bit integer)
  HWND  hwndMyConsoleWindowHandle;
  char  *pszMyConsoleWindowTitle;
  BOOL fShowControlPanel = FALSE; // preset options..
  BOOL fTestInput = FALSE;        // test OUTPUT if not otherwise specified
  BOOL fTestOutput= FALSE;
  BOOL fContinue  = TRUE;
  SYSTEMTIME windoze_system_time;

  printf("\r\nAudio-I/O-DLL Tester by DL4YHF. Compiled %s.\r\n",(char*)__DATE__);

  if( argc<2 )
   { printf("Missing argument. Usage: AudioIO_Test <dll_name> [options].\r\n" );
     printf("Options :\r\n /i = input test\r\n /o = output test");
     printf("\r\n /c = show control panel\r\n");
     printf("Example :\r\n");
     printf(" AudioIO_Test.exe c:\\cbproj\\AudioIO\\in_AudioIO.dll /o=VLF /i=VLF /c\r\n");
     printf("\r\nPress any key to continue.");
     getch();
     return -1;
   }


  AIO_Host_InitInstanceData( &AIO_Host );
  ChunkInfo_Init( &ChunkInfoForOutput );
  ChunkInfoForOutput.dblPrecSamplingRate = 44100.0; // default sampling rate for the OUTPUT
  ChunkInfoForOutput.dblSampleValueRange = 1.0;  // we use samples normalized to +/- 1.0 here
  ChunkInfoForOutput.dwValidityFlags =
              CHUNK_INFO_SAMPLE_RATE_VALID
     // ex: | CHUNK_INFO_SAMPLE_FORMAT_VALID  (removed; the 'sample format' must ALWAYS be valid)
            | CHUNK_INFO_SAMPLE_RATE_CALIBRATED
            | CHUNK_INFO_TIMESTAMPS_VALID; // applies to .rtRecTime, .dblUnixDateAndTime, .dblSecondsOfDay
  ChunkInfoForOutput.rtRecTime = 0.0;  // number of seconds since the start of recording / analysis / current track
  GetSystemTime( &windoze_system_time ); // win32 API . Should return UTC. "Seems to have a millisecond resolution". But no precision!
  ChunkInfoForOutput.ldblUnixDateAndTime = UTL_ConvertYMDhmsToUnix(
           windoze_system_time.wYear,  windoze_system_time.wMonth, windoze_system_time.wDay,
           windoze_system_time.wHour,  windoze_system_time.wMinute,
           (double)windoze_system_time.wSecond + 0.001*(double)windoze_system_time.wMilliseconds );
  // ex: ChunkInfoForOutput.dblSecondsOfDay = fmod( ChunkInfoForOutput.dblUnixDateAndTime, 24.0*60.0*60.0 );
  //     unnecessary, since dblUnixDateAndTime became ldblUnixDateAndTime (LONG DOUBLE)
  ChunkInfo_Init( &ChunkInfoFromInput );
  input_error = output_error = 0;

  // Try to load the default Audio-IO DLL .  We'll leave it to AudioIO.C
  //  to find the right directory, because we can only talk to Winamp
  //  if that DLL ("in_AudioIO.dll") is located in the winamp plugin directory.
  //  On a german PC, that would typically be "c:\Programme\Winamp\Plugins" .
  if( (i32Result=AIO_Host_LoadDLL(&AIO_Host, argv[1] )) < 0 )
   { printf("Couldn't load \"%s\",\r\n error description: %s .",
               (char*)argv[1],
               (char*)AIO_Host_ErrorCodeToString(&AIO_Host, i32Result) );
     printf("\r\nPress any key");
     getch();
     return -1;
   }
  else  // successfully loaded the special winamp-plugin; go and play...
   { printf("\r\nSuccessfully loaded the DLL (%s) .", (char*)argv[1] );
   }

  // Note that BOTH scenarios explained in aio2winamp.c, (A) and (B)
  //      as explained in AIO_OpenAudioOutput(), can be tested this way:
  //  AudioIO_Test_Client.exe c:\cbproj\AudioIO\in_AudioIO_BCB4.dll /c /i /o
  //  AudioIO_Test_Client.exe c:\cbproj\AudioIO\in_AudioIO_BCB4.dll /c /o /i
  //
  for(i=2; i<argc; ++i)  // parse the 'options' (argv[1]=DLL to be loaded)
   { if(strcmp(argv[i],"/c")==0)
        fShowControlPanel = TRUE;
     if(strncmp(argv[i],"/i",2)==0)
      { fTestInput = TRUE;
        if(argv[i][2]=='=')
         { strncpy(AioTest_sz40InputStreamID, argv[i]+3, 40 );
         }
        OpenAudioInput();
      } // end if < test INPUT ? >
     if(strncmp(argv[i],"/o",2)==0)
      { fTestOutput = TRUE;
        if(argv[i][2]=='=')
         { strncpy(AioTest_sz40OutputStreamID, argv[i]+3, 40 );
         }
        OpenAudioOutput();
      } // end if < test OUTPUT ? >
   }

  // What would Winamp "like to have" from us, and is it running at all ?
  AIO_Host_GetWinampParams( &AIO_Host,
              &dblWinampSampleRate,&iWinampNumChannels,
              &iWinampInitialized, &iWinampInputPaused );
  printf("\r\nWinamp: SR=%lg, %d channels, initialized=%d, paused=%d. ",
        (double)dblWinampSampleRate, (int)iWinampNumChannels,
           (int)iWinampInitialized,  (int)iWinampInputPaused );
      // Note: you should launch Winamp, and select in_AudioIO as input plugin,
      //       before starting this simple test app.
      // If the line above prints "initialized=0", there may be something wrong
      // with the shared memory in the DLL. Because from window's point of view,
      // Winamp and this test applications are two applications which shouldn't
      // know anything about each other.....    more details in aio2winamp.c !

  // Try to open the Audio-I/O-DLL's own control panel ?
  if( fShowControlPanel )
   {
        // Trick to retrieve the console window's handle,
        // found at http://www.flounder.com/consoledialogs.htm (here simplified)
        pszMyConsoleWindowTitle = "AudioIO_Test_Cient";
        SetConsoleTitle( pszMyConsoleWindowTitle );
        hwndMyConsoleWindowHandle = FindWindow(NULL, pszMyConsoleWindowTitle);
        i32Result=AIO_Host_ShowControlPanel( // NON-MODAL (not blocking) !
                           &AIO_Host, // [in] DLL-host instance data
           hwndMyConsoleWindowHandle, // [in] handle of the host's window ("parent")
                   &hwndControlPanel, // [out] handle of the DLL's control panel
            AIO_DISPLAY_SHOW_WINDOW,
                     (void*)&AIO_Host , // [in,optional] pvUserDataForCallback
            AIO_Host_Callback4ControlPanel ); // [in,optional] user callback function
        if(i32Result==AIO_NO_ERROR)
         {
           printf("\r\nSuccessfully opened the control panel.");
           HandleEvents();
         }
        else
         { printf("\r\nCouldn't open the control panel: %s",
           (char*)AIO_Host_ErrorCodeToString( &AIO_Host, i32Result ) );
         }
   } // end if( fShowControlPanel )

  iHadWaited_ms = 0;
  nSamplesRead  = 0;

  if( fTestOutput || fTestInput )
   { printf("\r\nProcessing 30*%d audio samples,",(int)L_SAMPLES_PER_BUFFER);
     if( fTestInput && fTestOutput )
          printf(" input+output .");
     else if( fTestInput )
          printf(" input only .");
     else printf(" output only .");
   }
  printf("\r\nLegend: .=successful input, -=no samples available, E=input error,\r\n"\
         "\to=successful output, O=output error (with numeric error code).\r\n");
  for(i=0; (i<30) || fContinue; ++i)
      { if( i==30 )
         { printf("\r\nClose control panel, or press ESC to finish.");
           if( fTestInput ) printf("\r\nPress 'i' to close input prematurely, 'I' to re-open, 'd' to dump chunk-info.");
           if( fTestOutput) printf("\r\nPress 'o' to close output prematurely, 'O' to re-open.");
           printf("\r\n");
         }

        if( fTestOutput ) // test the OUTPUT (at the same time as the input, or simultaneously)
         {
           dblPhaseAccuInc     = 2000.0/*Hz*/  * 2.0 * C_PI / ChunkInfoForOutput.dblPrecSamplingRate;
           dblFreqModPhaseAccuInc = 0.1/*Hz*/  * 2.0 * C_PI / ChunkInfoForOutput.dblPrecSamplingRate;
           dblFactor = 0.01; // 1.0 =  0 dBfs;  0.5 = -6 dBfs;  0.1 = -20 dBfs; 0.01 = -40 dBfs
           for(j=0; j<L_SAMPLES_PER_BUFFER; ++j)
            { dblFreqMod = 1000.0 * 2.0 * C_PI * cos( dblFMPhaseAccu );
              dblFMPhaseAccu += dblFreqModPhaseAccuInc;
              if( dblFMPhaseAccu > (2.0 * C_PI) )
               {  dblFMPhaseAccu -=(2.0 * C_PI);
               }
              // dblFreqMod = 0.0; // << test WITHOUT frequency modulation
              fltOutputBuffer[2*j+0] = dblFactor * sin( dblPhaseAccu + dblFreqMod );
              fltOutputBuffer[2*j+1] = dblFactor * cos( dblPhaseAccu + dblFreqMod );
              dblPhaseAccu += dblPhaseAccuInc;
              if( dblPhaseAccu > (2.0 * C_PI) )
               {  dblPhaseAccu -=(2.0 * C_PI);
               }
            }
           // To make things simple, a blocking call is used here (but "tamed"
           //  with a reliable timeout) when sending the samples to the output.
           //  A real-world application can avoid this by calling AIO_CanWrite(),
           //  to check *IF* and *HOW MANY* bytes can be written to the audio-output .
           //  (btw, the same principle is used in the Winamp "input" plugin) .
           //  Anyway, 8192 samples at 44.1 kHz take about 185 ms playing time,
           //  so 200 ms "timeout" should be ok here :
           ChunkInfoForOutput.dwNrOfSamplePoints = L_SAMPLES_PER_BUFFER;
           i32Result = AIO_Host_WriteOutputSamplePoints(
                 &AIO_Host,       // [in,out] DLL-host instance data
                 fltOutputBuffer, // [in] audio samples, as 32-bit FLOATING POINT numbers, grouped as "sample points"
                 L_SAMPLES_PER_BUFFER,  // [in] number of SAMPLE POINTS(!) to write or send
                 L_CHANNELS_PER_BUFFER, // [in] number of samples PER SAMPLE POINT
                 200,             // [in] max timeout in milliseconds, or AIO_TIMEOUT_NON_BLOCKING
                 &ChunkInfoForOutput, // [in,optional] chunk info, see c:\cbproj\SoundUtl\ChunkInfo.h, MAY BE NULL !
                 &iHaveWaited_ms); // [out,optional] "been-waiting-flag" (32-bit integer)
               iHadWaited_ms += iHaveWaited_ms;
               if( i32Result >= AIO_NO_ERROR )
                { printf("o");    // 'o' = successfully wrote OUTPUT
                  ChunkInfoForOutput.i64TotalSampleCounter += L_SAMPLES_PER_BUFFER; // not .. += i32Result !
                  dblChunkLengthInSeconds = (double)L_SAMPLES_PER_BUFFER/ChunkInfoForOutput.dblPrecSamplingRate;
                  ChunkInfoForOutput.rtRecTime += dblChunkLengthInSeconds;
                  ChunkInfoForOutput.ldblUnixDateAndTime += dblChunkLengthInSeconds;
                  // ChunkInfoForOutput.dblSecondsOfDay  += dblChunkLengthInSeconds;
                }
               else
                { printf("O%d ",(int)-i32Result);
                  output_error = i32Result;
                }
         } // end if( fTestOutput )

        if( fTestInput )
         { //  To keep things simple, we use a blocking call here
           //     (but "tamed" with a reliable timeout).
           //  A real-world application can avoid this by calling AIO_CanWrite(),
           //  to check *IF* and *HOW MANY* bytes can be written to the audio-output .
           //  (btw, the same principle is used in the Winamp "input" plugin) .
           //  8192 samples at 44.1 kHz take about 185 ms playing time,
           //  so 200 ms "timeout" should be ok here :
           i32Result = AIO_Host_ReadInputSamplePoints(
                 &AIO_Host,       // [in,out] DLL-host instance data
                 fltInputBuffer, // [in] audio samples, as 32-bit FLOATING POINT numbers, grouped as "sample points"
                 L_SAMPLES_PER_BUFFER,  // [in] number of SAMPLE POINTS(!) to read
                 L_CHANNELS_PER_BUFFER, // [in] number of samples PER SAMPLE POINT (may be less than provided by source)
                 200,             // [in] max timeout in milliseconds, or AIO_TIMEOUT_NON_BLOCKING
                 &ChunkInfoFromInput, // [out,optional] chunk info, see c:\cbproj\SoundUtl\ChunkInfo.h
                 &iHaveWaited_ms); // [out,optional] "been-waiting-flag" (32-bit integer)
           iHadWaited_ms += iHaveWaited_ms;
           if( i32Result >= AIO_NO_ERROR )
             { if( i32Result > 0 )
                { nSamplesRead += i32Result;
                  printf(".");    // '.' = successfully read from INPUT
                }
               else  // no error, but no samples available:
                { printf("-");
                }
             }
           else
             { printf("E%d ",(int)-i32Result);
               input_error = i32Result;
             }
         } // end if < test INPUT >

        HandleEvents();
        Sleep(20);   // Slow down the "filling" to see the buffer usage indicator in action
        if(fShowControlPanel && fControlPanelClosed)
         { fContinue = FALSE;
         }
        if( kbhit() )
         { switch( getch() )
            { case 27:
                 fContinue = FALSE;
                 break;
              case 'i':
                 AIO_Host_CloseAudioInput( &AIO_Host );
                 break;
              case 'I':
                 OpenAudioInput();
                 break;
              case 'o':
                 AIO_Host_CloseAudioOutput( &AIO_Host );
                 break;
              case 'O':
                 OpenAudioOutput();
                 break;
              case 'd':
                 DumpChunkInfo( &ChunkInfoFromInput );
                 break;
              default:
                 break;
            }
         } // end if kbhit()
      } // end for < X blocks >

     if( fTestInput )   // show summary / results of the INPUT test ?
      { if( input_error )  // something went wrong while sending audio. What ?
         { printf("\r\n Error while reading audio: %s",
              (char*)AIO_Host_ErrorCodeToString( &AIO_Host, input_error ) );
         }
        printf("\r\n ReadInput has been waiting for a total of %d ms.",(int)iHadWaited_ms);
        printf("\r\n Total number of sample points read = %ld",(long)nSamplesRead );
        AIO_Host_CloseAudioInput( &AIO_Host );
      } // end if( fTestInput )

     HandleEvents();
     Sleep(20);   // Slow down the "filling" to see the buffer usage indicator in action


     if( fTestOutput )
      { if( output_error )  // something went wrong while sending audio. What ?
         { printf("\r\n Error while sending audio: %s",
              (char*)AIO_Host_ErrorCodeToString( &AIO_Host, output_error ) );
         }
        printf("\r\n WriteOutput has been waiting for a total of %d ms.",(int)iHadWaited_ms);
        AIO_Host_CloseAudioOutput( &AIO_Host );
      } // end if ( fTestOutput )

     if( fShowControlPanel && !fControlPanelClosed )
      {
        printf("\r\nWaiting for the Audio-I/O control panel to close.. please close it");
           // A few hints about using windows in a CONSOLE APPLICATION:
           // - http://www.flounder.com/consoledialogs.htm
           //   (contains the SetConsoleTitle/FindWindow trick to retrieve the hWindow)
           // - http://www.codeproject.com/KB/winsdk/winconsole.aspx?msg=3606149#xx3606149xx
           //   (uses an extra thread to use a dialog in a console app)
           // - http://www.mvps.org/user32/modal.html
           //   (has a very simple 'message pump' for dialog windows)
         while( HandleEvents() )  // handles all pending WINDOWS events
          {  // (kind of "message pump", doesn't belong in a CONSOLE APPLICATION but...
            if( fControlPanelClosed )
             { printf("\r\nControl panel closed.");
               break;
             }
#          if(1)
            Sleep(50);     // Inefficient waste of CPU power, but this is just a TEST PROGRAM..
#          else
            WaitMessage(); // WaitMessage() would wait FOREVER when there's nothing to do (so it's useless here)
#          endif
          }
       } // end if < graphic control panel still open >

  printf("\r\nAbout to free (unload) the DLL...");
  AIO_Host_FreeDLL( &AIO_Host );  // unload the DLL (only do this once, when terminating)
    // 2012-11-22 : Crashed in AIO_Host_FreeDLL() when unloading ExtIO_RTL.DLL .
    
  printf("\r\nSurvived unloading the DLL.");
  printf("\r\nPress any key to continue.");
  getch();


  return 0;
} // end main() / main2()

/* Leave an empty line after this for certain compilers ! */
